import numpy as np
from scipy.linalg import blas 
from scipy.linalg import lapack 
def dscal(alpha,v):
    """v=alpha*v"""
    for i,vi in enumerate(v):
        v[i]=alpha*vi
    return v

def ddot(x,y):
    """s=x^Tv"""
    res=0
    for xi,yi in zip(x,y):
        res+=xi*yi
    return res

def daxpy(x,y,a=1):
    "y=a*x+y"
    for i in range(len(x)):
        y[i]+=a*x[i]
    return y

def dger(alpha,x,y,a=None):
    "a=a+alpha*x*y^T"
    if a is None:
        a=np.zeros((len(x),len(y)))
    for i,xi in enumerate(x):
        for j,yj in enumerate(y):
            a[i,j]+=alpha*xi*yj
    return a
 

def dgemv(alpha,A,x,beta=0,y=None):
    """y=alpha*Ax+beta*y"""
    if y is None:
        y=np.zeros(A.shape[0])
    for i in range(A.shape[0]):
        y[i]*=beta
        for j in range(A.shape[1]):
            y[i]+=alpha*A[i,j]*x[j]
    return y

# ou
def dgemv(alpha,A,x,beta=0,y=None):
    """y=alpha*Ax+beta*y"""
    if y is None:
        y=np.zeros(A.shape[0])
    for i in range(A.shape[0]):
        y[i]*=beta
        y[i]+=alpha*A[i,:].dot(x)
    return y    

def dgemm(alpha,A,B,beta=0,C=None):
    """C=alpha*A*B+beta*C"""
    if C is None:
        C=np.zeros((A.shape[0],B.shape[1])) 
    for i in range(A.shape[0]):
        for j in range(B.shape[1]):
            C[i,j]*=beta
            for k in range(A.shape[1]):
                C[i,j]+=A[i,k]*B[k,j]
    return C
# ou
def dgemm(alpha,A,B,beta=0,C=None):
    """C=alpha*A*B+beta*C"""
    if C is None:
        C=np.zeros((A.shape[0],B.shape[1])) 
    for i in range(A.shape[0]):
        for j in range(B.shape[1]):
            C[i,j]*=beta
            C[i,j]+=A[i,:].dot(B[:,j])
    return C    
    
    
def diag_sv(A,b):
    """x=A\b, A diagonal"""
    x=np.zeros(A.shape[0])
    for k in range(A.shape[0]):
        x[k]=b[k]/A[k,k]
    return x

def tri_inf_sv(L,b):
    """x=L\b, L tri inf"""
    x=np.zeros(L.shape[0])    
    x[0]=b[0]/L[0][0];
    for k in range(1,L.shape[0]):
        x[k]=b[k]
        for j in range(k):
            x[k]-=L[k,j]*x[j]
        x[k]=x[k]/L[k,k]
    return x

def pivot(A,b):
    """Pivot de Gauss"""
    n=A.shape[0]
    for k in range(n-1):
        # colone k a mettre en echelon
        for i in range(k+1,n):
            for j in range(k+1,n):
                A[i,j]-=A[k,j]*A[i,k]/A[k,k]
            # mise a jour b
            b[i]-=A[i,k]*b[k]/A[k,k]
            A[i,k]=0 # pivot a 0        
    return A,b

## test dscal
#x=np.ones(10)
#
#dscal(2,x)
#print(x)
#blas.dscal(2,x)
#print(x)
#
## test dot
#print(ddot(x,x))
#print(blas.ddot(x,x))
#print(x.dot(x))
#
## daxpy
#x=np.ones(10)
#y=np.arange(10)
#
#daxpy(x,y,a=1)
#print(y)
#y=np.arange(10)*1.0
#print(blas.daxpy(x,y,a=1))
#print(y)



# dger
x=np.ones(10)
y=np.arange(10)
A=np.zeros((10,10))
dger(1,x,y,A)
print(A)
A=np.zeros((10,10))
A=blas.dger(1,x,y,a=A)
print(A)


#dgemv
A=np.ones((10,5))
x=np.arange(5)
y=np.ones(10)
y=A.dot(x)+y
print(y)
y=np.ones(10)
dgemv(1,A,x,1,y)
print(y)
y=np.ones(10)
blas.dgemv(1,A,x,1,y,overwrite_y=True)
print(y)

#dgemm
A=np.random.rand(4,3)
B=np.random.rand(3,5)
C=np.ones((4,5))

C=A.dot(B)+C
print(C)
C=np.ones((4,5))
dgemm(1,A,B,1,C)
print(C)
C=np.ones((4,5))*1.0
blas.dgemm(1,A,B,1,C,overwrite_c=True)
print(C)


#%%

x0=np.ones(5)
L=np.random.rand(5,5)
for i in range(5):
    for j in range(i+1,5):
        L[i,j]=0
b=L.dot(x0)


x=tri_inf_sv(L,b)

x00=np.linalg.solve(L,b)
x2=b.copy()


x3=blas.dtrsm(1,L,x2,lower=True)


#%%

x0=np.ones(5)
A=np.random.rand(5,5)
b=A.dot(x0)

pivot(A,b)
x=np.linalg.solve(A,b)

x0=np.ones(5)
A=np.random.rand(5,5)
b=A.dot(x0)
res=lapack.dgesv(A,b)
